package org.processmining.niektax.temporalrelations.model;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.niektax.temporalrelations.model.dissimilarity.ChiSquareTestDissimilarity;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityMetric;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityMetrics;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult;
import org.processmining.niektax.temporalrelations.model.dissimilarity.ExactBinomialTestDissimilarity;
import org.processmining.niektax.temporalrelations.model.dissimilarity.FishersExactTestDissimilarity;
import org.processmining.niektax.temporalrelations.model.dissimilarity.HeuristicDissimilarity;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;

public class TemporalRelationUtils {
	
	public static Map<String, Integer> getCounts(XLog log, String cnameConnector){
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		Map<String, Integer> counts = new HashMap<String, Integer>();
		for(XTrace trace : log){
			for(XEvent event : trace){
				String classifiedName = xce.extractName(event)+cnameConnector+xle.extractTransition(event);
				Integer currentOccurrence = counts.get(classifiedName);
				counts.put(classifiedName, currentOccurrence==null ? 1 : currentOccurrence+1);
			}
		}
		return counts;
	}
	
	public static Set<String> getClassifiedNames(XLog log, String cnameConnector){
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		
		Set<String> classifiedNames = new HashSet<String>();

		for(XTrace trace : log){
			for(XEvent event : trace){
				String classifiedName = xce.extractName(event)+cnameConnector+xle.extractTransition(event);
				classifiedNames.add(classifiedName);
			}
		}
		
		return classifiedNames;
	}
	
	public static Map<String, Table<String, String, Integer>> calculateLogRelations(XLog log, String cnameConnector){
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		
		Map<String, Table<String, String, Integer>> logRelationMap = new HashMap<String, Table<String, String, Integer>>();
		
		Table<String, String, Integer> df = HashBasedTable.create();
		Table<String, String, Integer> ef = HashBasedTable.create();
		Table<String, String, Integer> ep = HashBasedTable.create();
		Table<String, String, Integer> cooc = HashBasedTable.create();
		
		for(XTrace t : log){
			String lastActivity = null;
			Set<String> activitiesSeen = new HashSet<String>();
			Map<String, Set<String>> recordedMap = new HashMap<String, Set<String>>();
			Map<String, Integer> activityCounts = new HashMap<String, Integer>();
			for(XEvent xe: t){
				String classifiedName = xce.extractName(xe)+cnameConnector+xle.extractTransition(xe);
				Integer currentCounts = activityCounts.get(classifiedName) == null ? 1 : activityCounts.get(classifiedName)+1;
				activityCounts.put(classifiedName, currentCounts);

				if(lastActivity!=null){
					Integer currentDF = df.get(classifiedName, lastActivity);
					df.put(classifiedName, lastActivity, currentDF==null ? 1 : currentDF+1);
					
					Set<String> recordedFor = recordedMap.get(classifiedName);
					for(String seenActivity:activitiesSeen){
						Integer currentEF = ef.get(classifiedName, seenActivity);
						ef.put(classifiedName, seenActivity, currentEF==null ? 1 : currentEF+1);
						if(recordedFor==null)
							recordedFor = new HashSet<String>();
						if(!recordedFor.contains(seenActivity)){
							Integer currentEP = ep.get(seenActivity, classifiedName);
							ep.put(seenActivity, classifiedName, currentEP==null ? 1 : currentEP+1);
							recordedFor.add(seenActivity);
							recordedMap.put(classifiedName, recordedFor);
						}
					}
				}
				lastActivity = classifiedName;
				activitiesSeen.add(classifiedName);
			}
			
			for(String activity1 : activityCounts.keySet()){
				Integer count = activityCounts.get(activity1);
				for(String activity2 : activityCounts.keySet()){
					if(activity1.equals(activity2)){
						if(count>1){
							Integer currentCooc = cooc.get(activity1, activity2);
							cooc.put(activity2, activity1, currentCooc==null ? count-1 : currentCooc+count-1);
						}
					}else{
						Integer currentCooc = cooc.get(activity1, activity2);
						cooc.put(activity2, activity1, currentCooc==null ? count : currentCooc+count);
					}
				}
			}
		}
		
		Set<String> classifiedNames = getClassifiedNames(log, cnameConnector);
		for(String cn1 : classifiedNames){
			for(String cn2 : classifiedNames){
				if(df.get(cn1, cn2)==null)
					df.put(cn1, cn2, 0);
				if(ef.get(cn1, cn2)==null)
					ef.put(cn1, cn2, 0);
				if(ep.get(cn1, cn2)==null)
					ep.put(cn1, cn2, 0);
				if(cooc.get(cn1, cn2)==null)
					cooc.put(cn1, cn2, 0);
			}
		}
		
		logRelationMap.put("df", df);
		logRelationMap.put("ef", ef);
		logRelationMap.put("ep", ep);
		logRelationMap.put("cooc", cooc);
		
		return logRelationMap;
	}
	
	public static Set<DissimilarityResult> calculateEventuallyActivityDissimilarity(XLog splittedLog, String a1, String a2, String originalAb, DissimilarityMetrics dm, double thresholdP, double thresholdOdds, String cnameConnector){
		Map<String, Table<String, String, Integer>> logRelationMap = calculateLogRelations(splittedLog, cnameConnector);
		
		Table<String, String, Integer> ef = HashBasedTable.create();//logRelationMap.get("ef");
		Table<String, String, Integer> ep = logRelationMap.get("ep");
		Table<String, String, Integer> emptyTable = HashBasedTable.create();
		
		Map<String, Integer> counts = getCounts(splittedLog, cnameConnector);
		Set<String> classifiedNames = getClassifiedNames(splittedLog, cnameConnector);
		
		DissimilarityMetric d = null;
		switch(dm){
			case BINOMIAL_TEST : d = new ExactBinomialTestDissimilarity(a1, a2, originalAb, emptyTable, ef, ep, emptyTable, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case CHI_SQUARE_TEST : d = new ChiSquareTestDissimilarity(a1, a2, originalAb, emptyTable, ef, ep, emptyTable, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case HEURISTIC : d = new HeuristicDissimilarity(a1, a2, originalAb, emptyTable, ef, ep, emptyTable, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case FISHER_TEST : d = new FishersExactTestDissimilarity(a1, a2, originalAb, emptyTable, ef, ep, emptyTable, classifiedNames, counts, thresholdP, thresholdOdds); break;
			default : System.err.println("NO DISSIMILARITY METRIC SET!"); break;
		}
		return d.getDissimilarity();
	}
	
	public static Set<DissimilarityResult> calculateActivityDissimilarity(XLog splittedLog, String a1, String a2, String originalAb, DissimilarityMetrics dm, double thresholdP, double thresholdOdds, String cnameConnector){
		Map<String, Table<String, String, Integer>> logRelationMap = calculateLogRelations(splittedLog, cnameConnector);
		
		Table<String, String, Integer> df = logRelationMap.get("df");
		Table<String, String, Integer> ef = logRelationMap.get("ef");
		Table<String, String, Integer> ep = logRelationMap.get("ep");
		Table<String, String, Integer> cooc = logRelationMap.get("cooc");
		
		Map<String, Integer> counts = getCounts(splittedLog, cnameConnector);
		Set<String> classifiedNames = getClassifiedNames(splittedLog, cnameConnector);
		
		DissimilarityMetric d = null;
		switch(dm){
			case BINOMIAL_TEST : d = new ExactBinomialTestDissimilarity(a1, a2, originalAb, df, ef, ep, cooc, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case CHI_SQUARE_TEST : d = new ChiSquareTestDissimilarity(a1, a2, originalAb, df, ef, ep, cooc, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case HEURISTIC : d = new HeuristicDissimilarity(a1, a2, originalAb, df, ef, ep, cooc, classifiedNames, counts, thresholdP, thresholdOdds); break;
			case FISHER_TEST : d = new FishersExactTestDissimilarity(a1, a2, originalAb, df, ef, ep, cooc, classifiedNames, counts, thresholdP, thresholdOdds); break;
			default : System.err.println("NO DISSIMILARITY METRIC SET!"); break;
		}
		
		return d.getDissimilarity();
	}

}
